/** @file cgenfunct.h convenience functions for cGenerator */

/* FAU Discrete Event Systems Library (libfaudes)

   Copyright (C) 2006  Bernd Opitz
   Copyright (C) 2007  Thomas Moor
   Exclusive copyright is granted to Klaus Schmidt

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2.1 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA */

#ifndef CGENFUNCT_H
#define CGENFUNCT_H


#include "definitions.h"
#include "cgenerator.h"
#include "project.h"
#include "parallel.h"
#include "supcon.h"


namespace faudes {


/**
 * Controllability 
 *
 * Tests whether the candidate supervisor h is controllable w.r.t.
 * the plant g; see IsControllable
 *
 * @param rPlantGen
 *   Plant g generator
 * @param rSupCandGen
 *   Supervisor candidate h generator 
 *
 * @exception Exception
 *   Alphabets of generators don't match (id 500)
 *   Plant generator nondeterministic (id 501)
 *   Specification generator nondeterministic (id 503)
 *   Plant & Spec generator nondeterministic (id 504)
 *
 * @return 
 *   true / false
 */
bool cIsControllable(
  const cGenerator& rPlantGen, 
  const vGenerator& rSupCandGen);



/**
 * Nonblocking Supremal Controllable Sublanguage 
 *
 * Computes the supremal sublanguage of the specification h that is controllable
 * w.r.t. the plant g. The result is given as a trim deterministic generator that 
 * may be used to supervise g in order to enforce h.
 *
 * See "C.G CASSANDRAS AND S. LAFORTUNE. Introduction to Discrete Event 
 * Systems. Kluwer, 1999." for base algorithm.
 *
 * This implementation does not require h to be a sublanguage of g. However,
 * both generators must be deterministic and share the same alphabet.
 * Controllable events are extracted from the plant parameter, and copied
 * to the result.
 *
 *
 * @param rPlantGen
 *   Plant cGenerator
 * @param rSpecGen
 *   Specification Generator
 * @param rResGen
 *   Reference to resulting Generator, the
 *   minimal restrictive nonblocking supervisor
 *
 * @exception Exception
 *   Alphabets of generators don't match (id 100)
 *   plant nondeterministic (id 201)
 *   spec nondeterministic (id 203)
 *   plant and spec nondeterministic (id 204)
 */
void cSupConNB(
  const cGenerator& rPlantGen, 
  const vGenerator& rSpecGen, 
  cGenerator& rResGen);


/** 
 * Supremal Controllable Sublanguage.
 * 
 * Finds the "largest" sublanguage of h for that language of g is
 * controllable with respect to h. Note that the language of h is not
 * rtequired to be a a sublanguage of g, but both languages must share 
 * the same alphabet.
 *
 * See "C.G CASSANDRAS AND S. LAFORTUNE. Introduction to Discrete Event 
 * Systems. Kluwer, 1999." for base algorithm.
 *
 * This version creates a "reverse composition map" given as reference
 * parameter.
 * 
 * @param rPlantGen
 *   Plant Generator 
 * @param rSpecGen
 *   Specification Generator 
 * @param rResGen
 *   Reference to resulting Generator, the
 *   minimal restrictive supervisor
 *
 * @exception Exception
 *   Alphabets of generators don't match (id 100)
 *   plant nondeterministic (id 201)
 *   spec nondeterministic (id 203)
 *   plant and spec nondeterministic (id 204)
 */
void cSupCon(
  const cGenerator& rPlantGen, 
  const vGenerator& rSpecGen, 
  cGenerator& rResGen);



/**
 * Parallel composition of accessible states of generators rGen1 and rGen2.
 * This version will set the resulting generators controllability attributes to
 * represent the union of controllable events of the arguments. When FAUDES_CHECKD
 * is defined, consistency of controllability flags in the argiments is tested.
 * 
 * Computes symbolic state names in rResGen, if symbolic state names are
 * enabled in rGen1 and rGen2 and rResGen
 *
 * @param rGen1
 *   First generator
 * @param rGen2
 *   Second generator
 * @param rResGen
 *   Reference to resulting parallel composition generator
 */
void cParallel(const cGenerator& rGen1, const cGenerator& rGen2, cGenerator& rResGen);


/**
 * Minimized (deterministic) projection. 
 * Calls projectnondet, determine and statemin. This version passes on
 * controllability attributes from the argument.
 *
 * @param rGen
 *   Reference to generator
 * @param rProjectAlphabet
 *   Projection alphabet
 * @param rResGen
 *   Reference to resulting deterministic generator
 */
void cProject(const cGenerator& rGen, const EventSet& rProjectAlphabet, cGenerator& rResGen);


} // namespace faudes

#endif
