/** @file op_observercomputation.h 

Methods to compute natural projections that exhibit the obsrver property.
The observer algorithm is elaborated in
K. C. Wong and W. M. Wonham, "On the Computation of Observers in Discrete Event
Systems," Discrete Event Dynamic Systems, vol. 14, no. 1, pp. 55-107, 2004. 
In addition, methods to compute natural projections that exhibit
output control consistency (OCC) and local control consistency (LCC) are provided. See for example
K. Schmidt and C. Breindl, "On Maximal Permissiveness of Hierarchical and Modular Supervisory
Control Approaches for Discrete Event Systems," Workshop on Discrete Event Systems, 2008. 
*/

/* FAU Discrete Event Systems Library (libfaudes)

   Copyright (C) 2006  Bernd Opitz
   Exclusive copyright is granted to Klaus Schmidt

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2.1 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA */

#ifndef OP_OBSERVERVERCOMPUTATION_H

#include "libfaudes.h"
#include "op_debug.h"
#include "op_bisimulation.h"
#include <map>
#include <vector>
#include <stack>



using namespace faudes;
using namespace std;

namespace faudes {
	/**
	* Lm-observer computation.
	* This function modifies a given generator and an associated natural projection
	* such that the resulting natural projection is an Lm-observer for the language marked by 
	* the resulting generator. 
	* This function evaluates the observer algorithm as described in 
	* K. C. Wong and W. M. Wonham, "On the Computation of Observers in Discrete Event Systems," 
	* Discrete Event Dynamic Systems, vol. 14, no. 1, pp. 55-107, 2004. 
	*
	* The alphabet rHighAlph has to be a subset of the alphabet of rGenObs.
	* rGenObs must be a deterministic generator.
	* There are no further restrictions on parameters.
	*
	* @param rGenObs
	*    Low-level generator. It is modified by the algorithm by relabeling transitions and events
	* @param rHighAlph
	*    Initial abstraction alphabet
	* @param rNewHighAlph
	*    Modified abstraction alphabet such that the abstraction is an Lm-observer
	* @param rMapRelabeledEvents
	*    Maps the original events to sets of newly introduced events (accumulatoive, call clear before)
	*
	* <h4>Example: Computation of an Lm-observer</h4>
	* <table border=0> <tr> <td> <table>
	* <tr> <td> Original generator </td>  </tr>
	* <tr>
	* <td> @image html ex_observer_all.png </td>
	* </tr>
	* <tr>
    * <td> Original high-level alphabet (rHighAlph): alpha, beta, gamma </td> </tr>
	* </table> </td> </tr> <tr> <td> <table width=100%>
	* <tr> <td> Result of calcAbstAlphObs(rGenObs, rHighAlph, rNewHighAlph, rMapRelabeledEvents); </td> </tr>
	* <tr> <td> New high-level alphabet (rNewHighAlph): alpha, beta, gamma, aNewHLevent_1, eNewHLevent_1, hNewHLevent_1 </td> </tr>
	* <tr> <td> @image html ex_synthesis_observer_result.png </td> </tr>
	* </table> </td> </tr> </table>
 	*
	* @ingroup ObserverPlugin
	*/
	void calcAbstAlphObs(cGenerator& rGenObs, EventSet& rHighAlph, EventSet& rNewHighAlph, map<Idx, set<Idx> >&  rMapRelabeledEvents);

   /**
	* Lm-observer computation.
	* This function is called by calcAbstAlphObs(cGenerator& rGenObs, EventSet& rHighAlph, EventSet& rNewHighAlph, map<Idx, set<Idx> >&  rMapRelabeledEvents).
	* It modifies a given generator and an associated natural projection
	* such that the resulting natural projection is an Lm-observer for the language marked by 
	* the resulting generator. 
	* This function evaluates the observer algorithm as described in 
	* K. C. Wong and W. M. Wonham, "On the Computation of Observers in Discrete Event Systems," 
	* Discrete Event Dynamic Systems, vol. 14, no. 1, pp. 55-107, 2004. 
	*
	* the alphabets rHighAlph and rControllableEvents have to be subsets of the alphabet of rGenObs
	* rGenObs must be a deterministic generator
	* no further restrictions on parameters.
    *
    * @param rGenObs
    *    Low-level generator. It is modified by the algorithm by relabeling transitions and events
    * @param rControllableEvents
    *    Set of controllable events 
    * @param rHighAlph
    *    Initial abstraction alphabet
    * @param rNewHighAlph
    *    Modified abstraction alphabet such that the abstraction is an Lm-observer
    * @param rMapRelabeledEvents
    *    Maps the original events to sets of newly introduced events (accumulatoive, call clear before)
    */
    void calcAbstAlphObs(vGenerator& rGenObs, EventSet& rControllableEvents, EventSet& rHighAlph, EventSet& rNewHighAlph, map<Idx, set<Idx> >&  rMapRelabeledEvents);

   /**
	* Lm-observer computation.
	* This function is called by void calcAbstAlphObs(vGenerator& rGenObs, EventSet& rControllableEvents, EventSet& rHighAlph, EventSet& rNewHighAlph, map<Idx, set<Idx> >&  rMapRelabeledEvents).
	* It modifies a given generator and an associated natural projection
	* such that the resulting natural projection is an Lm-observer for the language marked by 
	* the resulting generator. 
	* This function evaluates the observer algorithm as described in 
	* K. C. Wong and W. M. Wonham, "On the Computation of Observers in Discrete Event Systems," 
	* Discrete Event Dynamic Systems, vol. 14, no. 1, pp. 55-107, 2004. 
	*
	* The alphabets rHighAlph and rControllableEvents have to be subsets of the alphabet of rGenObs.
	* rGenObs must be a deterministic generator.
	* There are no further restrictions on parameters.
    *
    * @param rGenObs
    *    Low-level generator. It is modified by the algorithm by relabeling transitions and events
    * @param rControllableEvents
    *    Set of controllable events 
    * @param rHighAlph
    *    Initial abstraction alphabet
    * @param rNewHighAlph
    *    Modified abstraction alphabet such that the abstraction is an Lm-observer
    * @param rMapChangedTrans
    *    Maps the original relabeled transitions to the new events
    */
    void calcAbstAlphObs(vGenerator& rGenObs, EventSet& rControllableEvents, EventSet& rHighAlph, EventSet& rNewHighAlph, map<Transition,Idx>& rMapChangedTrans);

   /**
    * Computation of the dynamic system for an Lm-observer.
    * This function computes the dynamic system that is needed for evaluating the observer 
    * algorithm.
    *
    * The alphabet rHighAlph has to be a subset of the alphabet of rGen.
	* rGen must be a deterministic generator.
	* There are no further restrictions on parameters.
	*
    * @param rGen
    *    Generator for which the dynamic system is computed
    * @param rHighAlph
    *    Abstraction alphabet
    * @param rGenDyn
    *    Generator representing the dynamic system
    */
    void calculateDynamicSystemObs(const vGenerator& rGen, EventSet& rHighAlph, vGenerator& rGenDyn);

	/**
	* Lm-observer computation including output control consistency (OCC).
	* This function modifies a given generator and an associated natural projection
	* such that the resulting natural projection is an Lm-observer for the language marked by 
	* the resulting generator and at the same time fulfills the output control consistency 
    * condition (OCC). 
	* This function evaluates the observer algorithm as described in 
	* K. C. Wong and W. M. Wonham, "On the Computation of Observers in Discrete Event Systems," 
	* Discrete Event Dynamic Systems, vol. 14, no. 1, pp. 55-107, 2004. 
	* with an extension to OCC as indicated in
	* K. Schmidt and C. Breindl, "On Maximal Permissiveness of Hierarchical and Modular Supervisory
	* Control Approaches for Discrete Event Systems," Workshop on Discrete Event Systems, 2008. 
	*
	* The alphabet rHighAlph has to be a subset of the alphabet of rGenObs.
	* rGenObs must be a deterministic generator.
	* There are no further restrictions on parameters.
	*
	* @param rGenObs
	*    Low-level generator. It is modified by the algorithm by relabeling transitions and events
	* @param rHighAlph
	*    Initial abstraction alphabet
	* @param rNewHighAlph
	*    Modified abstraction alphabet such that the abstraction is an Lm-observer
	* @param rMapRelabeledEvents
	*    Maps the original events to sets of newly introduced events (accumulatoive, call clear before)
	*
	* <h4>Example: Computation of an Lm-observer with output control consistency (OCC)</h4>
	* <table border=0> <tr> <td> <table>
	* <tr> <td> Original generator </td>  </tr>
	* <tr>
	* <td> @image html ex_observer_all.png </td>
	* </tr>
	* <tr>
    * <td> Original high-level alphabet (rHighAlph): alpha, beta, gamma </td> </tr>
	* </table> </td> </tr> <tr> <td> <table width=100%>
	* <tr> <td> Result of calcAbstAlphObsOCC(rGenObs, rHighAlph, rNewHighAlph, rMapRelabeledEvents); </td> </tr>
	* <tr> <td> New high-level alphabet (rNewHighAlph): alpha, beta, gamma, d, f, h, aNewHLevent_3, bNewHLevent_2, 
    * cNewHLevent_2, eNewHLevent_3 </td> </tr>
	* <tr> <td> @image html ex_synthesis_occ_result.png </td> </tr>
	* </table> </td> </tr> </table>
	*
	* @ingroup ObserverPlugin
	*/
	void calcAbstAlphObsOCC(cGenerator& rGenObs, EventSet& rHighAlph, EventSet& rNewHighAlph, map<Idx,set<Idx > > &  rMapRelabeledEvents);

   /**
	* Lm-observer computation including output control consistency (OCC).
	* This function is called by calcAbstAlphObsOCC(cGenerator& rGenObs, EventSet& rHighAlph, EventSet& rNewHighAlph, map<Idx,set<Idx > > &  rMapRelabeledEvents).
    * It modifies a given generator and an associated natural projection
	* such that the resulting natural projection is an Lm-observer for the language marked by 
	* the resulting generator and at the same time fulfills the output control consistency 
    * condition (OCC). 
	* This function evaluates the observer algorithm as described in 
	* K. C. Wong and W. M. Wonham, "On the Computation of Observers in Discrete Event Systems," 
	* Discrete Event Dynamic Systems, vol. 14, no. 1, pp. 55-107, 2004. 
	* with an extension to OCC as indicated in
	* K. Schmidt and C. Breindl, "On Maximal Permissiveness of Hierarchical and Modular Supervisory
	* Control Approaches for Discrete Event Systems," Workshop on Discrete Event Systems, 2008. 
	*
	* The alphabets rHighAlph and rControllableEvents have to be subsets of the alphabet of rGenObs.
	* rGenObs must be a deterministic generator.
	* There are no further restrictions on parameters.
    *
    * @param rGenObs
    *    Low-level generator. It is modified by the algorithm by relabeling transitions and events
    * @param rControllableEvents
    *    Set of controllable events 
    * @param rHighAlph
    *    Initial abstraction alphabet
    * @param rNewHighAlph
    *    Modified abstraction alphabet such that the abstraction is an Lm-observer
    * @param rMapChangedTrans
    *    Maps the original relabeled transitions to the new events
    */
	void calcAbstAlphObsOCC(vGenerator& rGenObs, EventSet& rControllableEvents, EventSet& rHighAlph, EventSet& rNewHighAlph, map<Transition,Idx>& rMapChangedTrans);

   /**
    * Computation of the dynamic system for an Lm-observer that also fulfills output control consistency (OCC).
    * This function computes the dynamic system that is needed for evaluating the observer 
    * algorithm with the additional OCC condition.
    *
    * The alphabet rHighAlph has to be a subset of the alphabet of rGen.
	* rGen must be a deterministic generator.
	* There are no further restrictions on parameters.
	*
    * @param rGen
    *    Generator for which the dynamic system is computed
    * @param rControllableEvents
    *    Set of controllable events
    * @param rHighAlph
    *    Abstraction alphabet
    * @param rGenDyn
    *    Generator representing the dynamic system
    */
	void calculateDynamicSystemObsOCC(const vGenerator& rGen, EventSet& rControllableEvents, EventSet& rHighAlph, vGenerator& rGenDyn);

	/**
	* Forward reachability algorithm to determine locally reachable marked states.
	* This function is called by calculateDynamicSystemObsOCC(const vGenerator& rGen, EventSet& rControllableEvents, EventSet& rHighAlph, vGenerator& rGenDyn);
	* It inserts transitions with a certain "mLabel" into the dynamic system by employing forward reachability.
	*
	* There are no restrictions on parameters.
	* 
	* @param rGen
	*	Generator for which the dynamic system is computed
	* @param rHighAlph
	*     Abstraction alphabet
	* @param lowState
	*	Start state for forward reachability
	* @param mLabel
	*	Event label for transitions to marked states
	* @param rGenDyn
	*     Generator representing the dynamic system
	*/
	void forwardReachabilityObs(const vGenerator& rGen, const EventSet& rHighAlph, Idx lowState, Idx mLabel, vGenerator& rGenDyn);

	/**
	* Backward reachability algorithm for the dynamic system compuptation that fulfills the observer 
	* and output control consistency (OCC) condition. 
	* This recursive function is called by calculateDynamicSystemObsOCC(const vGenerator& rGen, EventSet& rControllableEvents, EventSet& rHighAlph, vGenerator& rGenDyn).
    * It evaluates a backward local reachability procedure to support the computation 
	* of a dynamic system for computing a natural projection that is an Lm-observer and OCC. 
	* 
	* There are no restrictions on parameters.
	*
	* @param rTransSetX2EvX1
	*    TransSet sorted by X2-Ev-X1
	* @param rControllableEvents
	*    Set of controllable events
	* @param rHighAlphabet
	*    Abstraction alphabet
	* @param exitState
    	*    Index of the exit state under investigation
    	* @param currentState
    	*    Index of the current state under investigation
    	* @param controllablePath
    	*    Boolean value that is false if the current path is uncontrollable and true otherwise
    	* @param rExitLocalStatesMap
    	*    Map from the current exit state to its local predecessor states
    	* @param rDoneStates
    	*    States that have already been investigated
	*/
	void backwardReachabilityObsOCC(const TransSetX2EvX1& rTransSetX2EvX1, const EventSet& rControllableEvents, const EventSet& rHighAlphabet, Idx exitState, Idx currentState, bool controllablePath, map<Idx, map<Idx, bool> >& rExitLocalStatesMap, StateSet& rDoneStates);

	/**
	* Lm-observer computation including local control consistency (LCC).
	* This function modifies a given generator and an associated natural projection
	* such that the resulting natural projection is an Lm-observer for the language marked by 
	* the resulting generator and at the same time fulfills the local control consistency 
      * condition (LCC). 
	* This function evaluates the observer algorithm as described in 
	* K. C. Wong and W. M. Wonham, "On the Computation of Observers in Discrete Event Systems," 
	* Discrete Event Dynamic Systems, vol. 14, no. 1, pp. 55-107, 2004. 
	* with an extension to LCC as indicated in
	* K. Schmidt and C. Breindl, "On Maximal Permissiveness of Hierarchical and Modular Supervisory
	* Control Approaches for Discrete Event Systems," Workshop on Discrete Event Systems, 2008. 
	*
	* The alphabet rHighAlph has to be a subset of the alphabet of rGenObs.
	* rGenObs must be a deterministic generator.
	* There are no further restrictions on parameters.
	*
	* @param rGenObs
	*    Low-level generator. It is modified by the algorithm by relabeling transitions and events
	* @param rHighAlph
	*    Initial abstraction alphabet
	* @param rNewHighAlph
	*    Modified abstraction alphabet such that the abstraction is an Lm-observer
	* @param rMapRelabeledEvents
	*    Maps the original events to sets of newly introduced events (accumulatoive, call clear before)
	*
	* <h4>Example: Computation of an Lm-observer with local control consistency (LCC)</h4>
	* <table border=0> <tr> <td> <table>
	* <tr> <td> Original generator </td>  </tr>
	* <tr>
	* <td> @image html ex_observer_all.png </td>
	* </tr>
	* <tr>
    * <td> Original high-level alphabet (rHighAlph): alpha, beta, gamma </td> </tr>
	* </table> </td> </tr> <tr> <td> <table width=100%>
	* <tr> <td> Result of calcAbstAlphObsLCC(rGenObs, rHighAlph, rNewHighAlph, rMapRelabeledEvents); </td> </tr>
	* <tr> <td> New high-level alphabet (rNewHighAlph): alpha, beta, gamma, d, f, h, aNewHLevent_2, bNewHLevent_1, 
    * cNewHLevent_1, eNewHLevent_2, hNewHLevent_2 </td> </tr>
	* <tr> <td> @image html ex_synthesis_lcc_result.png </td> </tr>
	* </table> </td> </tr> </table>
	*
	* @ingroup ObserverPlugin
	*/
    void calcAbstAlphObsLCC(cGenerator& rGenObs, EventSet& rHighAlph, EventSet& rNewHighAlph, map<Idx,set<Idx > > &  rMapRelabeledEvents);

   /**
	* Lm-observer computation including local control consistency (LCC).
	* This function is called by calcAbstAlphObsLCC(cGenerator& rGenObs, EventSet& rHighAlph, EventSet& rNewHighAlph, map<Idx,set<Idx > > &  rMapRelabeledEvents).
    	* It modifies a given generator and an associated natural projection
	* such that the resulting natural projection is an Lm-observer for the language marked by 
	* the resulting generator and at the same time fulfills the local control consistency 
    	* condition (LCC). 
	* This function evaluates the observer algorithm as described in 
	* K. C. Wong and W. M. Wonham, "On the Computation of Observers in Discrete Event Systems," 
	* Discrete Event Dynamic Systems, vol. 14, no. 1, pp. 55-107, 2004. 
	* with an extension to LCC as indicated in
	* K. Schmidt and C. Breindl, "On Maximal Permissiveness of Hierarchical and Modular Supervisory
	* Control Approaches for Discrete Event Systems," Workshop on Discrete Event Systems, 2008. 
	*
	* The alphabets rHighAlph and rControllableEvents have to be subsets of the alphabet of rGenObs.
	* rGenObs must be a deterministic generator.
	* There are no further restrictions on parameters.
    	*
    	* @param rGenObs
    	*    Low-level generator. It is modified by the algorithm by relabeling transitions and events
    	* @param rControllableEvents
    	*    Set of controllable events 
    	* @param rHighAlph
    	*    Initial abstraction alphabet
    	* @param rNewHighAlph
    	*    Modified abstraction alphabet such that the abstraction is an Lm-observer
    	* @param rMapChangedTrans
    	*    Maps the original relabeled transitions to the new events
    	*/
    void calcAbstAlphObsLCC(vGenerator& rGenObs, EventSet& rControllableEvents, EventSet& rHighAlph, EventSet& rNewHighAlph, map<Transition,Idx>& rMapChangedTrans);

   	/**
    	* Computation of the dynamic system for an Lm-observer that also fulfills local control consistency (LCC).
    	* This function computes the dynamic system that is needed for evaluating the observer 
    	* algorithm with the additional LCC condition.
    	*
    	* The alphabet rHighAlph has to be a subset of the alphabet of rGen.
	* rGen must be a deterministic generator.
	* There are no further restrictions on parameters.
	*
    	* @param rGen
    	*    Generator for which the dynamic system is computed
    	* @param rControllableEvents
    	*    Set of controllable events
    	* @param rHighAlph
    	*    Abstraction alphabet
    	* @param rGenDyn
    	*    Generator representing the dynamic system
    	*/
	void calculateDynamicSystemObsLCC(const vGenerator& rGen, EventSet& rControllableEvents, EventSet& rHighAlph, vGenerator& rGenDyn);
	
	/**
	* Backward reachability algorithm for the dynamic system compuptation that fulfills the observer 
	* and local control consistency (LCC) condition. 
	* This recursive function is called by calculateDynamicSystemObsLCC(const vGenerator& rGen, EventSet& rControllableEvents, EventSet& rHighAlph, vGenerator& rGenDyn).
    	* It evaluates a backward local reachability procedure to support the computation 
	* of a dynamic system for computing a natural projection that is an Lm-observer and LCC. 
	* 
	* There are no restrictions on parameters.
	*
	* @param rTransSetX2EvX1
	*    TransSet sorted by X2-Ev-X1
	* @param rControllableEvents
	*    Set of controllable events
	* @param rHighAlphabet
	*    Abstraction alphabet
	* @param exitState
    	*    Index of the exit state under investigation
    	* @param currentState
    	*    Index of the current state under investigation
    	* @param controllablePath
    	*    Boolean value that is false if the current path is uncontrollable and true otherwise
    	* @param rExitLocalStatesMap
    	*    Map from the current exit state to its local predecessor states
    	* @param rDoneStates
    	*    States that have already been investigated
	*/
    void backwardReachabilityObsLCC(const TransSetX2EvX1& rTransSetX2EvX1, const EventSet& rControllableEvents, const EventSet& rHighAlphabet, Idx exitState, Idx currentState, bool controllablePath, map<Idx, map<Idx, bool> >& rExitLocalStatesMap, StateSet& rDoneStates);

   /**
    * Relabeling algorithm for the computation of an Lm-observer.
    * This function checks the termination criterion of the observer algorithm. If required, transitions of
    * the input generator are relabeled.
    *
    * The alphabets rHighAlph and rControllableEvents have to be subsets of the alphabet of rGenRelabel.
	* There are no further restrictions on parameters.
    *
    * @param rGenRelabel
    *    Generator whose transitions are modified
    * @param rControllableEvents
    *    Set of controllable events
    * @param rHighAlph
    *    Abstraction alphabet
    * @param rNewPartitions
    *    Vector containing the indices of the equivalence classes that partition the state space of rGenRelabel
    * @param rMapStateToPartition
    *    Maps each state of rGenRelabel to an equivalence class that must also be contained in rNewPartitions
    * @param rMapChangedTransReverse
    *    Maps the relabeled transitions to the original transitions
    * @param rMapChangedTrans
    *    Maps the the modified original transitions to its new events
    * @param rMapRelabeledEvents
    *    Maps the original events to the set of new events they were relabeled with
    */
    bool relabel(vGenerator& rGenRelabel, EventSet& rControllableEvents, EventSet& rHighAlph, vector<Idx>& rNewPartitions, map<Idx,Idx>& rMapStateToPartition, map<Transition,Transition>& rMapChangedTransReverse, map<Transition,Idx>& rMapChangedTrans, map<Idx, EventSet>& rMapRelabeledEvents);


   /**
    * Convenience function for relabeling events in a given generator.
    * This function inserts new events and respective transitions given by a relableing map
    * into a given generator. The function is used to adjust plant components to the rtelableing
    * from another plant component.
    *
	* There are no restrictions on parameters.
    *
    * @param rGenPlant
    *    Plant component automaton
    * @param rMapRelabeledEvents
    *    Maps the original events to sets of newly introduced events
    * @param rNewEvents
    *    returns the newly inserted events (accumulative, call clear before)
    */
    void insertRelabeledEvents(cGenerator& rGenPlant, const map<Idx, set<Idx> >&  rMapRelabeledEvents, cEventSet& rNewEvents);

   /**
    * Convenience function for relabeling events in a given generator.
    * This function inserts new events and respective transitions given by a relableing map
    * into a given generator. The function is used to adjust plant components to the rtelableing
    * from another plant component.
    *
	* There are no restrictions on parameters.
    *
    * @param rGenPlant
    *    Plant component automaton
    * @param rMapRelabeledEvents
    *    maps the original events to sets of newly introduced events
    */
    void insertRelabeledEvents(cGenerator& rGenPlant, const map<Idx, set<Idx> >&  rMapRelabeledEvents);

}


#define OP_OBSERVERCOMPUTATION_H
#endif 
 
