------------------------------------------------------------------------------
-- benchmark_philosophers_monolithic.lua --
--   compute monolithic nonblocking supervisor for n forks and n philosophers 
--   with k intermediate states for the WODES'08 "dining philosophers"
--   benchmark example.
-- The supervisor is written to "tmp_monolithic_supervisor_n_k.gen".
------------------------------------------------------------------------------

  --################################################
  --############# configure parameters #############
  --################################################
  
  -- number of philosophers and forks
  n = 3
  
  -- number of intermediate states
  k = 4
  
  --###############################################
  --############# "include" functions #############
  --###############################################
  
  -- make function Philosopher() defined in generate_philosopher.lua
  -- available to our script
  dofile('generate_philosopher.lua')
  
  -- make function Fork() defined in generate_fork.lua
  -- available to our script
  dofile('generate_fork.lua')

  --###################################################################
  --########### compute monolithic nonblocking supervisor #############
  --###################################################################
  
  -- report to console
  print('***************************************************************')
  print('******* benchmark_philosophers_monolithic.lua *****************')
  print('***************************************************************')
  print('***')
  print(string.format('*** computing monolithic nonblocking supervisor for n=%d forks and n=%d philosophers',n,n))
  print(string.format('*** with k=%d intermediate states...',k))
  
  -- catch invalid numbers
  if(k<1) then print('*** ERROR: k has to be 1 at minimum!') end
  if(n<2) then print('*** ERROR: n has to be 2 at minimum!') end
  
  -- ************************************************************  
  -- *** compute compound plant of n philosophers and n forks ***
  
  -- initial step: Fork n || Philosopher 1
  plant=Generator.New()
  Parallel(Fork(n,1),Philosopher(1,k),plant)
  
  -- composition with all remaining forks and philosophers
  forkphil=Generator.New()
  for i=2,n do
       Parallel(Fork(i-1,i),Philosopher(i,k),forkphil)
       Parallel(plant,forkphil,plant)
  end
  forkphil:Clear()
  plant:Name("monolithic plant")
  
  -- report to console
  plant:StatWrite()
  
  -- ************************************************************  
  -- ******** compute unrestricitve specification Sigma* ********
  -- unrestricitve specification: Sigma*
  spec=Generator.New()
  FullLanguage(plant:Alphabet(),spec)
  
  -- ************************************************************  
  -- ****************** supervisor computation ******************
  sup=Generator.New()
  SupConNB(plant,spec,sup)
  plant:Clear()
  spec:Clear()
  
  -- prepare result
  sup:ClearStateNames()
  sup:Name(string.format('Philosophers: monolithic nonblocking supervisor, n=%d, k=%d',n,k))
  
  -- report to console
  sup:StatWrite()
  
  -- Write result
  sup:Write(string.format('tmp_MonolithicSupervisor_n%d_k%d.gen',n,k))
  
  -- Remark: it has been verified that sup is state minimal
  
  print('***')
  print('*** ...done.')
  print('***************************************************************')